// ========================================================================
// Copyright 2006-2007 Mort Bay Consulting Pty. Ltd.
// ------------------------------------------------------------------------
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ========================================================================

package org.mortbay.jetty.client;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;

import org.mortbay.io.Buffer;
import org.mortbay.io.Buffers;
import org.mortbay.io.ByteArrayBuffer;
import org.mortbay.io.Connection;
import org.mortbay.io.EndPoint;
import org.mortbay.jetty.AbstractGenerator;
import org.mortbay.jetty.HttpGenerator;
import org.mortbay.jetty.HttpHeaderValues;
import org.mortbay.jetty.HttpHeaders;
import org.mortbay.jetty.HttpParser;
import org.mortbay.jetty.HttpTokens;
import org.mortbay.jetty.HttpVersions;
import org.mortbay.util.IO;


/**
 *
 * @author Greg Wilkins
 * @author Guillaume Nodet
 */
class HttpConnection implements Connection
{
    HttpDestination _destination;
    EndPoint _endp;
    HttpGenerator _generator;
    HttpParser _parser;
    boolean _http11;
    Buffer _connectionHeader;
    Buffer _requestContentChunk;

    /* The current exchange waiting for a response */
    HttpExchange _exchange;  
    
    /* The pipeline of exchanged for this connection */
    private LinkedList<HttpExchange> _exchanges=new LinkedList<HttpExchange>();
    
    /* ------------------------------------------------------------ */
    HttpConnection(Buffers buffers,EndPoint endp,int hbs,int cbs)
    {
        _endp=endp;
        _generator=new HttpGenerator(buffers,endp,hbs,cbs);
        _parser=new HttpParser(buffers,endp,new Handler(),hbs,cbs);
    }

    /* ------------------------------------------------------------ */
    public HttpDestination getDestination()
    {
        return _destination;
    }

    /* ------------------------------------------------------------ */
    public void setDestination(HttpDestination destination)
    {
        _destination = destination;
    }

    /* ------------------------------------------------------------ */
    public void handle() throws IOException
    {
        System.err.println("HANDLE "+this+" by "+Thread.currentThread());
        
        // Get the next exchange
        // TODO really handle pipelines
        synchronized(this)
        {
            if (_exchange==null && _exchanges.size()>0)
                _exchange=_exchanges.removeFirst();
            if (_exchange!=null)
            {
                System.err.println("ex="+_exchange+" "+_exchange.getStatus());

                if (_exchange.getStatus()==HttpExchange.STATUS_WAITING_FOR_COMMIT)
                    commitRequest();
            }
        }
        
        int no_progress=0;
        while (_endp.isOpen())
        {
            try
            {
                long io=0;
                
                // Write as much of the request as possible
                // TODO handle piplelines
                if (!_generator.isComplete() && _generator.isCommitted())
                {
                    io=_generator.flush();
                    
                    if (!_generator.isComplete())
                    {
                        InputStream in = _exchange.getRequestContentSource();
                        if (in!=null)
                        {
                            if (_requestContentChunk==null || _requestContentChunk.length()==0)
                            {
                                _requestContentChunk=_exchange.getRequestContentChunk();
                                System.err.println("pushing request chunk");
                                if (_requestContentChunk!=null)
                                    _generator.addContent(_requestContentChunk,false);
                                else
                                    _generator.complete();
                                io+=_generator.flush();
                            }
                        }
                        else
                            _generator.complete();
                    }
                }

                // If we are not ended then parse available
                if (!_parser.isComplete()) 
                    io+=_parser.parseNext();

                if (io>0)
                    no_progress=0;
                else if (no_progress++>=2) 
                {
                    if (!_endp.isBlocking())
                    {
                        return;
                    }
                }
            }
            finally
            {
                if (_parser.isComplete() && _generator.isComplete())
                {  
                    // TODO handle pipelines
                    reset(true);
                    _exchange=null;
                }
            }
        }
    }
    
    /* ------------------------------------------------------------ */
    public boolean isIdle()
    {
        return _exchange == null;
    }

    /* ------------------------------------------------------------ */
    public EndPoint getEndPoint()
    {
        return _endp;
    }

    /* ------------------------------------------------------------ */
    public void send(HttpExchange ex) throws IOException
    {
        System.err.println("HttpConnection.send "+ex);
        
        synchronized(this)
        {
            ex.setStatus(HttpExchange.STATUS_WAITING_FOR_COMMIT);
            _exchanges.add(ex);
            handle();
        }
    }
        

    /* ------------------------------------------------------------ */
    private void commitRequest() throws IOException
    {
        System.err.println("COMMIT! ");
        synchronized(this)
        {
            if (_exchange.getStatus()!=HttpExchange.STATUS_WAITING_FOR_COMMIT)
                throw new IllegalStateException();
            
            _exchange.setStatus(HttpExchange.STATUS_SENDING_REQUEST);

            _generator.setVersion(_exchange._version);
            _generator.setRequest(_exchange._method, _exchange._uri);

            if (_exchange._requestContent != null)
            {
                _exchange._requestFields.putLongField(HttpHeaders.CONTENT_LENGTH, _exchange._requestContent.length());
                _generator.completeHeader(_exchange._requestFields, false);
                _generator.addContent(_exchange._requestContent, true);
            }
            else if (_exchange._requestContentSource != null)
            {
                _generator.completeHeader(_exchange._requestFields, false);
                int available = _exchange._requestContentSource.available();
                if (available>0)
                {
                    // TODO deal with any known content length
                    System.err.println("Get a chunk of content");
                    
                    // TODO reuse this buffer!
                    byte[] buf = new byte[available];
                    int length =_exchange._requestContentSource.read(buf);
                    _generator.addContent(new ByteArrayBuffer(buf,0,length), false);
                }
            }
            else
            {
                _exchange._requestFields.remove(HttpHeaders.CONTENT_LENGTH); // TODO: should not be needed
                _generator.completeHeader(_exchange._requestFields, true);
            }
            
            _generator.flush();
            _exchange.onRequestCommitted();

            if (_exchange instanceof HttpExchange.StreamedExchange)
            {
                ((HttpExchange.StreamedExchange) _exchange)._parser = _parser;
            }

            _exchange.setStatus(HttpExchange.STATUS_WAITING_FOR_RESPONSE);
            
            
            if (_endp.isBlocking())
            {
                handle();
            }
        }
    }

    /* ------------------------------------------------------------ */
    protected void reset(boolean returnBuffers) throws IOException
    {
        _connectionHeader = null;
        _parser.reset(returnBuffers); 
        _generator.reset(returnBuffers); 
        _http11=false;
        
        boolean close = shouldClose();
       
        synchronized(this)
        {
            if (_exchanges.size()==0)
                _destination.returnConnection(this, close);
        }
    }
    
    /* ------------------------------------------------------------ */
    private boolean shouldClose()
    {
        if (_connectionHeader == HttpHeaderValues.CLOSE_BUFFER)
        {
            return true;
        }
        else if (_connectionHeader == HttpHeaderValues.KEEP_ALIVE_BUFFER)
        {
            return false;
        }
        else
        {
            return !_http11;
        }
    }

    /* ------------------------------------------------------------ */
    private class Handler extends HttpParser.EventHandler
    {
        @Override
        public void startRequest(Buffer method, Buffer url, Buffer version) throws IOException
        {
            throw new IllegalStateException();
        }

        @Override
        public void startResponse(Buffer version, int status, Buffer reason) throws IOException
        {
            _http11 = (version == HttpVersions.HTTP_1_1_BUFFER);
            _exchange.onResponseStatus(version, status, reason);
            _exchange.setStatus(HttpExchange.STATUS_PARSING_HEADERS);
        }
        
        @Override
        public void parsedHeader(Buffer name, Buffer value) throws IOException
        {
            if (name == HttpHeaders.CONNECTION_BUFFER)
            {
                _connectionHeader = HttpHeaderValues.CACHE.lookup(value);
            }
            //_exchange._responseFields.add(name, value);
            _exchange.onResponsetHeader(name, value);
        }

        @Override
        public void headerComplete() throws IOException
        {
            _exchange.setStatus(HttpExchange.STATUS_PARSING_CONTENT);
            _exchange.onResponseHeaderComplete();
        }

        @Override
        public void content(Buffer ref) throws IOException
        {
            _exchange.onResponseContent(ref);
        }

        @Override
        public void messageComplete(long contextLength) throws IOException
        {
            _exchange.onResponseComplete();
            _exchange.setStatus(HttpExchange.STATUS_COMPLETED);
        }
    }
    
    public String toString()
    {
        return "HttpConnection@"+hashCode()+"//"+_destination.getAddress().getHostName()+":"+_destination.getAddress().getPort();
    }
    
}
